// credits: based on
//
// Catalano Machine Learning Library
// The Catalano Framework
//
// Copyright  Diego Catalano, 2013
// diego.catalano at live.com
//
// Copyright  Andrew Kirillov, 2007-2008
// andrew.kirillov@gmail.com
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//

package icsoc20;

import java.util.List;
import java.util.Map;
import java.util.Random;

import org.sat4j.specs.TimeoutException;

import de.ovgu.featureide.fm.core.base.IFeature;
import fr.univ.lille.MeasureFinder;
import fr.univ.lille.exploration.FMExplorer;
import fr.univ.lille.measurements.ConfigurationValidator;

/**
 * Implements the FM-guided exploration strategies. Follows similar flow as e-greedy, but replaces
 * random selection of next action with FM-guided (IncrementalStrategy or IncrementalDegreeStrategy)
 * selection of next action.
 * 
 * @author Andreas Metzger
 */
public class FMExploration implements IExplorationPolicy{
    private double epsilon; // exploration rate
    private double delta; // to introduce a little randomization to bootstrap FM-guided exploration
    
    private Random r = new Random();
    private Random q = new Random();

    private FMExplorer explorer = null;
    private MeasureFinder finder = null;
    
    private int maxAction;
    private int previousMaxAction = 0;
    
    private boolean evolAware = true;
    
    private Map<String,Integer> confAction;
    private Map<Integer,String> actionConf;
    
    /**
     * Initializes a new instance of the FMExploration class.
     * @param epsilon Epsilon value (exploration rate).
     * @param confAction A map that maps configurations to action numbers (connecting FM-guided and Q-Learning).
     * @param actionConf A map that maps action numbers to configurations (connecting FM-guided and Q-Learning).
     * @param strategy The FM-guided strategy to be used (2 = Incremental, 3 = IncrementalDegree)
     * @param maxAction Highest permissible action in action space (to scope action space during evolution)
     * @param evolAware If true, consider new part of action space first
     */
    public FMExploration(double epsilon,  Map<String,Integer> confAction, 
    		Map<Integer,String> actionConf, int strategy, int maxAction, boolean evolAware) {
        this.epsilon = epsilon;
        
        this.maxAction = maxAction;
        this.evolAware = evolAware;
        
        
        this.confAction = confAction;
        this.actionConf = actionConf;
        
		this.finder = new MeasureFinder(strategy, "resources/models/CloudRM.xml", "resources/measures/CloudRM.xml");
		finder.loadMeasures();
		
    }

    public double getEpsilon() {
        return epsilon;
    }

    /**
     * Set epsilon.
     * 
     * @param epsilon Epsilon value (exploration rate), [0, 1].
     */
    public void setEpsilon(double epsilon) {
        this.epsilon = Math.max( 0.0, Math.min( 1.0, epsilon ) );
    }
    

    /**
     * Set delta, to introduce a little randomization to bootstrap FM-guided exploration.
     * 
     * @param delta Delta value, [0, 1].
     */

    public void setDelta(double delta) {
        this.delta = Math.max( 0.0, Math.min( 1.0, delta ) );
    }

    public void setMaxAction(int maxAction) {
        this.maxAction = maxAction;
    }

    public void setPreviousMaxAction(int maxAction) {
        this.previousMaxAction = maxAction;
    }
    
    /**
     * Action selection by either exploiting knowledge or exploring the action space (using
     * FM-guided strategies)
     * 
     * @param actionEstimates Action Estimates (Q-Table).
     * @return Return Selected actions.
     */
    @Override
    public int ChooseAction(double[] actionEstimates){
        int actionsCount = maxAction;
    	
        // Exploitation: find the best action (greedy)
        double maxReward = actionEstimates[0];
        int greedyAction = 0;

        for ( int i = 0; i < actionsCount; i++ )
        {
            if ( actionEstimates[i] > maxReward )
            {
                maxReward = actionEstimates[i];
                greedyAction = i;
            }
        }

        // Exploration: find the action using FM-guided exploration
        if ( r.nextDouble( ) < epsilon )
        {
        	int fmAction = 0;
        	String fmConfigOld;

        	// use a certain amount of randomness at the start to ensure
        	// configuration space is "explored" and not stuck in local minima
        	if(q.nextDouble() < delta) {
        		fmAction = r.nextInt( actionsCount - 1 );
        	} else {
                // Find the next configuration starting from greedy action (exploitation)
        		
        		// (re-init) Strategy (to start search the whole space of configurations)
        		// NOTE: Current workaround for not having a setConfig method
        		this.explorer = finder.getExplorer();
        		
        		// Get config string from greedy action
        		String currentConfig = actionConf.get(greedyAction);
        		String nextConfig = "";
        		List<String> nextConfiguration;
        		
        		while(explorer.getExplorationStrategy().hasRemainingConfigurations()) {
        			nextConfiguration = explorer.getNextConfiguration();
        			String config = ConfigurationValidator.formatConfiguration(nextConfiguration);
        			
        			// remember this action, which may be used if new part of adaptation space is exhausted
        			// "our strategies first explore the feature combinations that were added to the 
        			// adaptation space by an evolution step, and then explore the remaining feature combinations 
        			// if needed"
        			fmConfigOld = config;
        			
        			if(config.equals(currentConfig)) {
        				if(explorer.getExplorationStrategy().hasRemainingConfigurations()) {
        					do {
	        					if(explorer.getExplorationStrategy().hasRemainingConfigurations()) {
	        						nextConfiguration = explorer.getNextConfiguration();
		        					nextConfig = ConfigurationValidator.formatConfiguration(nextConfiguration);
		        					fmAction = confAction.get(nextConfig);	
	        					} else {
	        						nextConfig = fmConfigOld;
	        						break;
	        					}
	        					// actions should be smaller than the maximum space
	        					if(fmAction >= actionsCount) continue;
        					} while((evolAware && fmAction < previousMaxAction));
    					
        				} else {
        					nextConfig = "-1";
        				}
        					
        			}
        		}

        		if(nextConfig.equals("")) {
        			System.out.println("\nERROR: Could not match starting config");
        			fmAction = greedyAction;
        		} else if(nextConfig.equals("-1")) {
        			// WARNING: Configuration space exhausted
        			System.out.print("W");
        			fmAction = greedyAction;        		
        		} else {
        			fmAction = confAction.get(nextConfig);
        		}
        		
        	}        	
        	return fmAction;
        }


        
        return greedyAction;
    }
    
}